--[[

   world.lua
   Generates the platform game map of Because it's fun, Fay.
	
   Copyright 2008 Oxeye Game Studio. http://www.oxeyegames.com
	
   File contributors:
      Jens Bergensten

   This game was written for the TIGSource Procedural Generation
   competition in May 2008. You may use this Lua source code in
   any way you please, we don't mind. However, the game's sound
   effects and artwork belong to us, so please ask for permission
   before you do anything funny =)

]]

local gameField = nil
local gameFieldRotated = nil
local gameFieldSprites = nil
local DIMA = 32  -- x width
local DIMB = 300 -- y height
local keys = nil
local jumpers = nil
local doors = nil
local doodads = nil
local windowSprite = nil

local function generateTilesetPattern(fieldCopy, gameFieldSprites, tiles, reset)
	-- nw, n, ne, w, e, sw, s, se
	local patternFinder = {
		{ 2, 4, 5, 7 },
		{ 2, 4, 7 },
		{ 2, 4, 5 },
		{ 2, 5, 7 },
		{ 4, 5, 7 },
		{ 2, 4 },
		{ 2, 5 },
		{ 5, 7 },
		{ 4, 7 },
		{ 4 },
		{ 2 },
		{ 5 },
		{ 7 },
		{ 1 },
		{ 3 },
		{ 8 },
		{ 6 }
	}
	local patternClearer = {
		{ 1, 2, 3, 4, 5, 6, 7, 8 },
		{ 1, 2, 3, 4, 6, 7, 8 },
		{ 1, 2, 3, 4, 5, 6, 8 },
		{ 1, 2, 3, 5, 6, 7, 8 },
		{ 1, 3, 4, 5, 6, 7, 8 },
		{ 1, 2, 3, 4, 6 },
		{ 1, 2, 3, 5, 8 },
		{ 3, 4, 6, 7, 8 },
		{ 1, 4, 6, 7, 8 },
		{ 1, 4, 6 },
		{ 1, 2, 3 },
		{ 3, 5, 8 },
		{ 6, 7, 8 },
		{ 1 },
		{ 3 },
		{ 8 },
		{ 6 }
	}
	local patternSprite = {
		1,
		14,
		15,
		16,
		17,
		10,
		11,
		12,
		13,
		2,
		3,
		4,
		5,
		6,
		7,
		8,
		9
	}

	for x=2,DIMA-1 do
		for y=2,DIMB-1 do
			local value = fieldCopy[x][y]
			if value > 0 then
				if reset then
					gameFieldSprites[x][y] = {}
				end
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(((value - 1) * 40) + 1, ((value - 1) * 40) + 6)]
			end

			-- scan neighbors more valued
			for i=value+1,8 do
				-- nw, n, ne, w, e, sw, s, se
				local bitField = { 0, 0, 0, 0, 0, 0, 0, 0 }

				if fieldCopy[x-1][y-1] == i then bitField[1] = 1 end
				if fieldCopy[x][y-1] == i then bitField[2] = 1 end
				if fieldCopy[x+1][y-1] == i then bitField[3] = 1 end
				if fieldCopy[x-1][y] == i then bitField[4] = 1 end
				if fieldCopy[x+1][y] == i then bitField[5] = 1 end
				if fieldCopy[x-1][y+1] == i then bitField[6] = 1 end
				if fieldCopy[x][y+1] == i then bitField[7] = 1 end
				if fieldCopy[x+1][y+1] == i then bitField[8] = 1 end

				for pattern=1,17 do
					local found = true
					for p=1,#(patternFinder[pattern]) do
						if bitField[ patternFinder[pattern][p] ] == 0 then
							found = false
							break
						end
					end
					if found then
						-- clear pattern
						for p=1,#(patternClearer[pattern]) do
							bitField[ patternClearer[pattern][p] ] = 0
						end
						-- place relevant sprite
						local s = (i - 1) * 40 + patternSprite[pattern] * 2 + 5
						gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(s,s+1)]
					end
				end
			end
		end
	end
end

local function placeBackgroundSprites(field, spriteField, metaField)

	local tiles = {}

	for i=1,320 do
		local spriteName = "0"
		if i < 10 then
			spriteName = spriteName .. "00" .. i
		elseif i < 100 then
			spriteName = spriteName .. "0" .. i
		else
			spriteName = spriteName .. i
		end
		tiles[i] = daisy.createSpriteState(spriteName, "bgTiles.dat")
	end

	-- make a copy of the level so we can modify it
	local fieldCopy = {}

	for x=1,DIMA do
		fieldCopy[x] = {}

		for y=1,DIMB do
			if field[x][y] == 1 then
				fieldCopy[x][y] = 1
			else
				fieldCopy[x][y] = 0
			end
		end
	end

	local shadowTileset = 8
	local roomTileset = 5
	local brickTileset = 4
	local dirtTileset = 1
	local overlayTileset = 6

	for x=2,DIMA-1 do
		for y=2,DIMB-1 do
			if y >= DIMB - 90 then
				fieldCopy[x][y] = dirtTileset
			else
				fieldCopy[x][y] = brickTileset
			end
		end
	end

	-- place some cleared rectangles
	local count = math.random(10, 50)
	for i=1,count do
		local left, top, right, bottom = math.random(2, DIMA - 1), math.random(2, DIMB - 1), math.random(2,6), math.random(2,6)
		right = right + left
		bottom = bottom + top

		if right > DIMA - 1 then
			right = DIMA - 1
		end
		if bottom > DIMB - 1 then
			bottom = DIMB - 1
		end

		for x=left,right do
			for y=top,bottom do
				fieldCopy[x][y] = 0
			end
		end
	end

	-- draw some "natural" stuff
	count = math.random(20, 100)
	for i=1,count do
		local cx,cy = math.random(2, DIMA - 1), DIMB - math.abs(math.random(0, 120) + math.random(0, 120) - 70)
		local left, top, right, bottom = cx - math.random(1, 9), cy - math.random(1, 9), cx + math.random(1,9), cy + math.random(1, 9)
		local type = math.random(2,3)

		if left < 2 then
			left = 2
		end
		if top < 2 then
			top = 2
		end
		if right > DIMA - 1 then
			right = DIMA - 1
		end
		if bottom > DIMB - 1 then
			bottom = DIMB - 1
		end
		local wMax = right - left + 1
		local hMax = bottom - top + 1

		for x=left,right do
			for y=top,bottom do
				local dx,dy = math.abs(x - cx) / wMax, math.abs(y - cy) / hMax

				if math.random() >= dx + dy then
					fieldCopy[x][y] = type
				end
			end
		end
	end

	-- place rooms
	for x=2,DIMA-1 do
		for y=2,DIMB-1 do
			if metaField[x][y] == 1 and metaField[x-1][y] == 1 and metaField[x+1][y] == 1 and metaField[x][y-1] == 1 and metaField[x][y+1] == 1 then
				fieldCopy[x][y] = roomTileset
			end
		end
	end

	-- the topmost rows must be cleared
	for x=1,DIMA do
		for y=1,8 do
			if y < 8 or (x % 4) ~= 3 then
				fieldCopy[x][y] = 0
			end
		end
	end

	generateTilesetPattern(fieldCopy, gameFieldSprites, tiles, false)
	
	-- test
	for x=1,DIMA do
		for y=1,DIMB do
			if fieldCopy[x][y] ~= roomTileset and fieldCopy[x][y] ~= brickTileset then
				fieldCopy[x][y] = 0
			end
		end
	end
	count = math.random(10, 50)
	for i=1,count do
		local cx,cy = math.random(2, DIMA - 1), DIMB - math.abs(math.random(0, 120) + math.random(0, 120) - 70)
		local left, top, right, bottom = cx - math.random(1, 7), cy - math.random(1, 7), cx + math.random(1,7), cy + math.random(1, 7)
		local type = math.random(2,3)

		if left < 2 then
			left = 2
		end
		if top < 2 then
			top = 2
		end
		if right > DIMA - 1 then
			right = DIMA - 1
		end
		if bottom > DIMB - 1 then
			bottom = DIMB - 1
		end
		local wMax = right - left + 1
		local hMax = bottom - top + 1

		for x=left,right do
			for y=top,bottom do
				if fieldCopy[x][y] ~= roomTileset then
					local dx,dy = math.abs(x - cx) / wMax, math.abs(y - cy) / hMax

					if math.random() >= dx + dy then
						fieldCopy[x][y] = type
					end
				end
			end
		end
	end
	
	-- place some overlay rectangles on the bricks
	count = math.random(10, 50)
	for i=1,count do
		local cx,cy = math.random(2, DIMA - 1), math.random(10, DIMB / 2)
		local left, top, right, bottom = cx - math.random(1, 3), cy - math.random(1, 3), cx + math.random(1,3), cy + math.random(1, 3)

		if left < 2 then
			left = 2
		end
		if top < 2 then
			top = 2
		end
		if right > DIMA - 1 then
			right = DIMA - 1
		end
		if bottom > DIMB - 1 then
			bottom = DIMB - 1
		end
		local wMax = right - left + 1
		local hMax = bottom - top + 1

		for x=left,right do
			for y=top,bottom do
				local dx,dy = math.abs(x - cx) / wMax, math.abs(y - cy) / hMax

				if math.random() >= dx + dy and
					fieldCopy[x][y] == brickTileset and
					fieldCopy[x-1][y] == brickTileset and
					fieldCopy[x+1][y] == brickTileset and
					fieldCopy[x][y-1] == brickTileset and
					fieldCopy[x][y+1] == brickTileset then
					
					fieldCopy[x][y] = overlayTileset
					
				end
			end
		end
	end
	
	-- erase rooms and bricks, we don't want them duplicated
	for x=1,DIMA do
		for y=1,DIMB do
			if fieldCopy[x][y] == roomTileset or fieldCopy[x][y] == brickTileset then
				fieldCopy[x][y] = 0
			end
		end
	end
	
	-- place shadows
	for x=1,DIMA do
		for y=1,DIMB do
			if field[x][y] == 1 then
				fieldCopy[x][y] = shadowTileset
			end
		end
	end
			
	generateTilesetPattern(fieldCopy, gameFieldSprites, tiles, true)

	-- now place vines
	for x=1,DIMA do
		for y=1,DIMB do
			if field[x][y] == -1 then
				fieldCopy[x][y] = 7 -- vine set
			else
				fieldCopy[x][y] = 0
			end
		end
	end
	
	generateTilesetPattern(fieldCopy, gameFieldSprites, tiles, false)
	
	-- place some windows
	count = math.random(15, 30)
	for i=1,count do
	
		local y = math.random(4, 30) * 4
		local x = math.random(1, 7) * 4
		
		if field[x][y] == 0 and
			field[x+1][y] == 0 and
			field[x-1][y] == 0 and
			field[x][y-1] == 0 and
			field[x+1][y-1] == 0 and
			field[x-1][y-1] == 0 and
			field[x][y+1] == 0 and
			field[x+1][y+1] == 0 and
			field[x-1][y+1] == 0 and
			field[x][y-2] == 0 and
			#(gameFieldSprites[x][y]) > 0 and
			#(gameFieldSprites[x][y-1]) > 0 then
			
			gameFieldSprites[x][y] = {}
			gameFieldSprites[x][y-1] = {}
			
			table.insert(doodads, { x = (x-1) * 32, y = (y - 1) * 32, sprite = windowSprite })
			--print("a")
			
		end
	end
	
end

local function placePlatformSprites(field, spriteField)

	local tiles = {}
	for i=1,80 do
		local spriteName = "00"
		if i < 10 then
			spriteName = spriteName .. "0" .. i
		else
			spriteName = spriteName .. i
		end
		tiles[i] = daisy.createSpriteState(spriteName, "tiles.dat")
	end

	-- make a copy of the level so we can modify it
	local fieldCopy = {}

	for x=1,DIMA do
		fieldCopy[x] = {}

		for y=1,DIMB do
			if field[x][y] == 1 then
				fieldCopy[x][y] = 1
			else
				fieldCopy[x][y] = 0
			end
			if y == DIMB then
				fieldCopy[x][y] = 0
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(41,48)]
			elseif x == 1 or x == DIMA then
				fieldCopy[x][y] = 0
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(1,8)]
			end
		end
	end

	-- scan for platforms
	-- 1 normal, 2 facing left, 3 facing right, 4 right edge, 5 left edge, 6 T section, 7 bottom edge, 8 middle run, 9 downright, 10 downleft, 11 upright, 12 upleft
	local x,y = 2, 2
	while y < DIMB do

		if fieldCopy[x][y] > 0 then
			if fieldCopy[x][y - 1] > 0 then
				if fieldCopy[x][y - 1] == 6 or fieldCopy[x][y - 1] == 8 then
					-- middle
					if fieldCopy[x][y + 1] > 0 then
						fieldCopy[x][y] = 8
					else
						fieldCopy[x][y] = 7
					end
				elseif fieldCopy[x][y - 1] == 2 or fieldCopy[x][y - 1] == 12 then
					-- facing left
					if fieldCopy[x][y + 1] > 0 then
						fieldCopy[x][y] = 2
					else
						fieldCopy[x][y] = 10
					end
				else
					-- facing right
					if fieldCopy[x][y + 1] > 0 then
						fieldCopy[x][y] = 3
					else
						fieldCopy[x][y] = 9
					end
				end
			elseif fieldCopy[x][y + 1] > 0 then
				if x == 2 or fieldCopy[x - 1][y] == 0 then
					fieldCopy[x][y] = 12
				elseif x == (DIMA-1) or fieldCopy[x + 1][y] == 0 then
					fieldCopy[x][y] = 11
				else
					fieldCopy[x][y] = 6
				end
			else
				if fieldCopy[x - 1][y] == 0 and x > 2 then
					fieldCopy[x][y] = 5
				elseif fieldCopy[x + 1][y] == 0 and x < DIMA-1 then
					fieldCopy[x][y] = 4
				end
			end
		end

		x = x + 1
		if x >= DIMA then
			y = y + 1
			x = 2
		end
	end

	for x=1,DIMA do
		for y=1,DIMB do
			if fieldCopy[x][y] == 1 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(1,8)]
			elseif fieldCopy[x][y] == 2 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(9,12)]
			elseif fieldCopy[x][y] == 3 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(13,16)]
			elseif fieldCopy[x][y] == 4 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(17,20)]
			elseif fieldCopy[x][y] == 5 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(21,24)]
			elseif fieldCopy[x][y] == 6 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(25,26)]
			elseif fieldCopy[x][y] == 7 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(27,28)]
			elseif fieldCopy[x][y] == 8 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(29,32)]
			elseif fieldCopy[x][y] == 9 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(33,34)]
			elseif fieldCopy[x][y] == 10 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(35,36)]
			elseif fieldCopy[x][y] == 11 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(37,38)]
			elseif fieldCopy[x][y] == 12 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = tiles[math.random(39,40)]
			end
		end
	end
end

local function createDoor(x, sy, floors, metaField)
	if sy < DIMB-3 and #floors > 0 then
		local kind = math.random(2,5)
		gameField[x][sy] = kind
		gameField[x][sy+1] = kind
		
		table.insert(doors, {kind=kind, x=x*32 - 16, y=sy*32, open=false, anim=0})

		-- create a key
		local done = false
		local ground = sy + 1
		repeat
			for index,floor in pairs(floors) do
				if floor.level > ground and math.random() < .25 then
					local keyX = floor.lineStart
					if floor.lineStart < floor.lineEnd then
						keyX = math.random(floor.lineStart, floor.lineEnd)
					end
					if gameField[keyX][floor.level] == 0 and metaField[keyX][floor.level] ~= 2 then
						done = true
						table.insert(keys, { posX = keyX * 32 - 16, posY = floor.level * 32 - 16, type = kind, angle=math.random()*math.pi })
						if metaField[keyX][floor.level] == 0 then
							metaField[keyX][floor.level] = 2
						end
						if metaField[keyX][floor.level+1] == 0 then
							metaField[keyX][floor.level+1] = 2
						end
						break
					end
				end
			end
		until done
	end
end

local function createRoom(sx, ex, currentFloor, floorHeight, floorLines, metaField)
	for i=1,floorHeight-3 do
		gameField[sx][currentFloor+i] = 1
		gameField[ex][currentFloor+i] = 1
	end
	if math.random() > .5 then
		createDoor(sx, currentFloor + floorHeight - 2, floorLines, metaField)
	end
	if math.random() > .5 then
		createDoor(ex, currentFloor + floorHeight - 2, floorLines, metaField)
	end

	-- mark this room
	for y=0,floorHeight do
		for x=sx,ex do
			metaField[x][currentFloor + y] = 1
		end
	end
end

local callRootFunction = nil
local topMostVine = DIMB
local function createVineStem(left, top, right, bottom, field, meta, y, sx, ex)

	local sy = y
	local spawnedChild = false
	while y > 20 do

		-- fill this row
		for x=sx,ex do
			if (meta[x][y] == 5 or meta[x][y] == 0) and field[x][y] == 0 then
				field[x][y] = -1
			end
		end

		if y < topMostVine then
			topMostVine = y
		end

		local nextStep = math.random(1,6)
		if nextStep == 1 then
			if sx < ex then
				sx = sx + 1
			end
		elseif nextStep == 2 then
			if sx > left then
				sx = sx - 1
			end
		elseif nextStep == 3 then
			if sx < ex then
				ex = ex - 1
			end
		elseif nextStep == 4 then
			if ex < right then
				ex = ex + 1
			end
		elseif nextStep == 5 and (((sy - y) >= 5 and math.random() > .75) or y <= bottom) then
			if not spawnedChild and y <= bottom then
				local temp = math.random()
				if temp > .5 then
					if (sx - left) >= 2 then
						callRootFunction(left, top, sx, y, field, meta)
					elseif (right - ex) >= 2 then
						callRootFunction(ex, top, right, y, field, meta)
					else
						--print("FAILED " .. left .. ", " .. right)
					end
				else
					if (right - ex) >= 2 then
						callRootFunction(ex, top, right, y, field, meta)
					elseif (sx - left) >= 2 then
						callRootFunction(left, top, sx, y, field, meta)
					else
						--print("FAILED " .. left .. ", " .. right)
					end
				end
			end
			break
		else
			if (ex - sx) == 6 then
				-- spawn a child
				spawnedChild = true
				createVineStem(left, top, right, bottom, field, meta, y, sx, sx + 1)
				sx = ex - 1
			end
		end

		if left > 2 and math.random() > .75 then
			left = left - 1
		end
		if right < (DIMA-1) and math.random() > .75 then
			right = right + 1
		end

		y = y - 1

	end

end

local function createVineRoot(left, top, right, bottom, field, meta)

	-- look for a suitable platform to begin at
	local y = bottom + 1
	while y < DIMB do
		local runStarted = false
		local sx = 0
		local ex = 0

		for x=left,right do
			if field[x][y] > 0 then
				if not runStarted then
					runStarted = true
					sx = x
				end
				ex = x
			else
				if runStarted and (ex - sx) >= 2 then
					-- good enough platform
					break
				end
				runStarted = false
			end
		end

		if runStarted and (ex - sx) >= 2 then
			sx = math.random(sx, ex - 2)
			ex = math.random(sx + 2, ex)
			createVineStem(left, top, right, bottom, field, meta, y, sx, ex)
			break
		end

		y = y + 1
	end
end
callRootFunction = createVineRoot

local function checkVineOrphan(sx, ex, floor, field, metaField)

	local leftOff = math.abs((sx-1) - (DIMA / 2))
	local rightOff = math.abs((ex+1) - (DIMA / 2))
	local left,top,right
	local vineX = sx - 1

	if sx <= 2 or rightOff < leftOff then
		left,top,right = ex + 1, floor, ex + 4
		if left > (DIMA-1) then
			left = (DIMA-1)
		end
		if right > (DIMA-1) then
			right = (DIMA-1)
		end
		if top > DIMB then
			top = DIMB
		end
		vineX = ex + 1
	else
		left,top,right = sx - 4, floor, sx - 1
		if left < 2 then
			left = 2
		end
		if right < 2 then
			right = 2
		end
		if top > DIMB then
			top = DIMB
		end
	end

	local scanLength = 1
	local found = false
	while (top + scanLength) < DIMB and not found do
		for x=left,right do
			if field[x][top + scanLength] ~= 0 then
				found = true
				break
			end
		end
		if not found then
			scanLength = scanLength + 1
		end
	end

	if scanLength >= 5 then
		--print("Orphan! " .. sx .. ", " .. ex .. ", " .. floor .. ", " .. scanLength)

		local height = math.random(2,4)
		if scanLength >= 5 then
			height = math.random(scanLength / 2, scanLength)
		end
		for y=floor,floor+height do
			if field[vineX][y] == 0 then
				field[vineX][y] = -1
			end
		end
	end
end

local function checkMetaCollision(left,top,right,bottom,field)
	for y=top,bottom do
		for x=left,right do
			if field[x][y] ~= 0 then
				return true
			end
		end
	end
	return false
end

local function makeGameField()
	gameField = {}
	local floorLines = {}
	local metaField = {} -- 0 nothing, 1 rooms, 2 doors, 3 stairs, 4 jumps, 5 open

	for x=1,DIMA do
		gameField[x] = {}
		metaField[x] = {}

		for y=1,DIMB do
			if x == 1 or x == DIMA or y == DIMB then
				gameField[x][y] = 1
			else
				gameField[x][y] = 0
			end
			metaField[x][y] = 0
		end
	end
	table.insert(floorLines, 1, { lineStart = 2, lineEnd = DIMA - 1, level = DIMB - 1 } )

	local y = DIMB - 4
	local prevSX = 4
	local prevEX = 20
	local lastStep = 4
	while y > 13 do

		local sx, ex
		local sContained, eContained
		repeat
			sx = math.random(3, 16)
			ex = math.random(sx + 2, DIMA - 2)
			sContained = (sx > prevSX and sx < prevEX)
			eContained = (ex > prevSX and ex < prevEX)
		until eContained or sContained

		for x=sx,ex do
			gameField[x][y] = 1
		end
		table.insert(floorLines, 1, { lineStart = sx, lineEnd = ex, level = y - 1 } )

		if sContained then
			if math.random() < .5 then
				-- run e to end
				for x=ex+1,DIMA do
					gameField[x][y] = 1
				end
				table.insert(floorLines, 1, { lineStart = ex + 1, lineEnd = DIMA - 1, level = y - 1 } )
				ex = DIMA
			end
			if math.random() < .4 and sx > 4 then
				-- second platform
				local newEx = math.random(2, sx - 3)
				for x=1,newEx do
					gameField[x][y] = 1
				end
				table.insert(floorLines, 1, { lineStart = 2, lineEnd = newEx, level = y - 1 } )
			end
			if lastStep >= 4 and ex >= (prevEX + 2) and sx < (prevEX - 2) then
				createRoom(sx, prevEX - 1, y, lastStep, floorLines, metaField)
			end
		else
			if math.random() < .5 then
				-- run s to start
				for x=1,sx-1 do
					gameField[x][y] = 1
				end
				table.insert(floorLines, 1, { lineStart = 2, lineEnd = sx-1, level = y - 1 } )
				sx = 1
			end
			if math.random() < .4 and ex < 29 then
				-- second platform
				local newSx = math.random(ex + 3, DIMA)
				for x=newSx,DIMA do
					gameField[x][y] = 1
				end
				table.insert(floorLines, 1, { lineStart = newSx, lineEnd = DIMA - 1, level = y - 1 } )
			end
			if lastStep >= 4 and sx <= (prevSX - 2) and ex > (prevSX + 2) then
				createRoom(prevSX + 1, ex, y, lastStep, floorLines, metaField)
			end
		end

		prevSX = sx
		prevEX = ex

		lastStep = math.random(3,5)
		y = y - lastStep
	end

	-- create open areas?
	y = math.random(20, DIMB/2)
	while y < (DIMB - 80) do
		local clearHeight = math.random(20, 50)

		for cy=y,y+clearHeight do
			for cx=2,DIMA-1 do
				if metaField[cx][cy] == 0 then
					gameField[cx][cy] = 0
					metaField[cx][cy] = 5
				end
			end
		end

		topMostVine = y + clearHeight
		while topMostVine > y do
			createVineRoot(2, y, DIMA - 1, topMostVine + 1, gameField, metaField)
		end

		-- look for orphaned platforms to prevent impossible maps
		local startHeight = topMostVine - 1
		if startHeight > (y - 3) then
			startHeight = (y - 3)
		end
		if startHeight <= 1 then
			startHeight = 2
		end
		for fy=startHeight,y+clearHeight do
			local sx, ex = 0, 0
			local runStarted = false
			local fx = 2

			while fx < (DIMA-1) do

				if gameField[fx][fy] == 1 and gameField[fx][fy-1] ~= 1 then
					if runStarted then
					else
						sx = fx
						runStarted = true
					end
					ex = fx
				else
					if runStarted then
						-- end of platform
						checkVineOrphan(sx, ex, fy, gameField, metaField)
						runStarted = false
					end
				end
				fx = fx + 1
			end
			if runStarted then
				checkVineOrphan(sx, ex, fy, gameField, metaField)
			end

		end

		y = y + clearHeight + math.random(30, 100)
	end

	-- attempt to make some jumper areas
	for key,line in pairs(floorLines) do
		if line.level > 10 and line.level < (DIMB-4) and (line.lineEnd - line.lineStart) >= 5 then

			local sx = math.random(line.lineStart, line.lineEnd - 4)
			local ex = sx + 4

			if not checkMetaCollision(sx, line.level - 8, ex, line.level+1, metaField) then

				-- clear...
				for my=line.level-7,line.level do
					for mx=sx,ex do
						metaField[mx][my] = 4
						metaField[mx][my+1] = 4
						metaField[mx][my-1] = 4
						gameField[mx][my] = 0
					end
				end

				-- place jumper
				table.insert(jumpers, { posX = (sx + 2) * 32, posY = line.level * 32 })
			end

		end
	end

	-- attempt to make some stair areas
	for key,line in pairs(floorLines) do
		if line.lineStart <= 2 or line.lineEnd >= (DIMA - 1) then
			-- this is a floor line next to the wall, check for space
			local stairSize = math.random(2,7)

			local sx,ex
			if line.lineStart <= 2 then
				sx,ex = 2,2 + stairSize
			else
				sx,ex = (DIMA-1) - stairSize,(DIMA-1)
			end

			local collision = sx < line.lineStart or ex > line.lineEnd or checkMetaCollision(sx, line.level-stairSize, ex, line.level, metaField)
			if not collision then
				for my=line.level-stairSize,line.level do
					for mx=sx,ex do
						metaField[mx][my] = 3
						metaField[mx][my+1] = 3
						metaField[mx][my-1] = 3
						gameField[mx][my] = 0
					end
				end
				if line.lineStart <= 2 then
					for step=0,stairSize do
						for mx=1,stairSize-step do
							gameField[1 + mx][line.level-step] = 1
						end
					end
				else
					for step=0,stairSize do
						for mx=1,stairSize-step do
							gameField[DIMA - mx][line.level-step] = 1
						end
					end
				end
			end
		end
	end

	gameFieldSprites = {}

	local sprites = {}
	sprites[-1] = daisy.createSpriteState("Vine", "gfx.dat")
	sprites[0] = daisy.createSpriteState("BgTile", "gfx.dat")
	sprites[1] = daisy.createSpriteState("PlatTile", "gfx.dat")
	--sprites[2] = daisy.createSpriteState("RedDoor", "gfx.dat")
	--sprites[3] = daisy.createSpriteState("GreenDoor", "gfx.dat")
	--sprites[4] = daisy.createSpriteState("PurpleDoor", "gfx.dat")
	--sprites[5] = daisy.createSpriteState("YellowDoor", "gfx.dat")

	for x=1,DIMA do
		gameFieldSprites[x] = {}

		for y=1,DIMB do
			gameFieldSprites[x][y] = {}
			--[[if gameField[x][y] == 1 then
				gameFieldSprites[x][y][1] = sprites[1]
			else
				gameFieldSprites[x][y][1] = sprites[0]
				if sprites[gameField[x][y] ] then
					gameFieldSprites[x][y][2] = sprites[gameField[x][y] ]
				end
			end
			]]
		end
	end
	placeBackgroundSprites(gameField, gameFieldSprites, metaField)
	placePlatformSprites(gameField, gameFieldSprites)

	for x=1,DIMA do
		for y=1,DIMB do
			if gameField[x][y] ~= 1 and gameField[x][y] ~= 0 then
				gameFieldSprites[x][y][#(gameFieldSprites[x][y]) + 1] = sprites[gameField[x][y] ]
			end
		end
	end

	gameFieldRotated = {}

	for x=1,DIMB do
		gameFieldRotated[x] = {}

		for y=1,DIMA do
			gameFieldRotated[x][y] = gameField[y][x]
		end
	end

	gameField.isPositionBlocked = function (x, y)
			if x < 1 or y < 1 or x > DIMA or y > DIMB then
				return true
			end
			return gameField[x][y] > 0
		end
	gameFieldRotated.isPositionBlocked = function (x, y)
			if x < 1 or y < 1 or x > DIMB or y > DIMA then
				return true
			end
			return gameFieldRotated[x][y] > 0
		end
end


function createWorld(keyObjects, jumpObjects, doorObjects, doods)

	keys = keyObjects
	jumpers = jumpObjects
	doors = doorObjects
	doodads = doods
	
	windowSprite = daisy.createSpriteState("window", "tiles.dat")

	makeGameField()

	return gameField,gameFieldRotated,DIMA,DIMB,gameFieldSprites
end